#include <iostream>
#include <cstdlib>
#include <cmath>
#include <ctime>
using namespace std;




#include <cuda_runtime_api.h>
#include <curand.h>
#include "curand_kernel.h"

#define DIMENSIONI 2
#define NPUNTI 100000000
#define THREADS 512
#define NBLOCKS 512



// init random numbers

__global__ void initRNG(curandState *  const rngStates, const unsigned int seed){

  unsigned int tid = blockIdx.x * blockDim.x + threadIdx.x;

  curand_init(seed, tid, 0, &rngStates[tid]);

}


////////////////////////////////////////////////////////////


__global__ void volumeSfera(curandState * const rngState, int *dev_c ){


  int tid = threadIdx.x;
  long int offset = blockDim.x*blockIdx.x + tid;

  double ran;  // per il generico random
  double r;    // per il raggio del generico punto

  __shared__ int ir[THREADS];



  curandState localState = rngState[offset];

  int temp = 0;

  while (offset< NPUNTI){

    r = 0;
    for(int i=0; i<DIMENSIONI; i++){
      //      ran = curand_uniform(&localState);
           ran = curand_uniform_double(&localState);
      r += pow(ran, 2);
    };

    if (sqrt(r) <= 1)
      temp +=1;

    offset += blockDim.x*gridDim.x;

  };
  ir[tid] = temp;

  rngState[offset] = localState;


  __syncthreads();


  int i = blockDim.x/2;
  while(i!=0){
    if(tid<i)
      ir[tid] +=ir[tid+i];

    __syncthreads();

    i /=2;
  }

  if(tid==0)
    dev_c[blockIdx.x]=ir[0];


};



////////////////////////////////////////////////////////////

int main(void){

  double integrale = 0;
  int *partial_c;
  int *dev_partial_c;
  long int sumint=0;





  partial_c = (int *)malloc(NBLOCKS*sizeof(int) );
  cudaMalloc( (void **)&dev_partial_c, NBLOCKS*sizeof(int) );



  dim3 block;
  dim3 grid;

  block.x = NBLOCKS;
  grid.x = THREADS;


  curandState *dev_rngState = 0;
  cudaMalloc( (void **)&dev_rngState, block.x * grid.x * sizeof(curandState) );

  int seed = 7777;
  initRNG<<<block,grid>>>( dev_rngState, seed);





  volumeSfera<<<NBLOCKS, THREADS>>>(dev_rngState, dev_partial_c);

  cudaMemcpy( partial_c, dev_partial_c, NBLOCKS * sizeof(int), cudaMemcpyDeviceToHost);

  double xpunti = (double)NPUNTI;
  for( int i=0; i <NBLOCKS; i++){
    integrale += (double)partial_c[i]/xpunti;
  }
  integrale = integrale*pow(2., DIMENSIONI);


  cout << "volume=" << integrale << "\n";


  cudaFree(dev_partial_c);
  free(partial_c);

  return 0;



};







